unit formDemo;

// (c) 2015 by Roy Magne Klever, rmklever@gmail.com, www.rmklever.com
// You are granted free usage as long as some credit is given
//
// This demo require ImageEn installed: www.imageen.com

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ComCtrls, ieview, imageenview, ExtCtrls, ieopensavedlg,
  hyieutils, GraphUtil, hyiedefs, imageenproc, Math;

type
  TProgressEvent = procedure(Progress: Integer; Text: String; var Cancel: Boolean) of Object;
  PRGB24 = ^TRGB24;
  TRGB24 = record B, G, R: Byte; end;
  PRGBArray = ^TRGBArray;
  TRGBArray = array[0..32767] of TRGB24;

  TfrmFx = class(TForm)
    pnlCtrls: TPanel;
    imgView: TImageEnView;
    tbRadius: TTrackBar;
    lblRadius: TLabel;
    btnOpen: TButton;
    btnSave: TButton;
    dlgOpen: TOpenImageEnDialog;
    dlgSave: TSaveImageEnDialog;
    btnPreview: TButton;
    shpPrg: TShape;
    procedure btnSaveClick(Sender: TObject);
    procedure btnOpenClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure btnPreviewClick(Sender: TObject);
    procedure tbRadiusChange(Sender: TObject);
  private
    procedure DoFX;
    procedure DoProgress(Progress: Integer; Text: string; var Cancel: Boolean);
    { Private declarations }
  public
    { Public declarations }
     TheImage, Img: TIEBitmap;
     ImgLoaded: Boolean;
  end;

var
  frmFx: TfrmFx;
  fxProgress: TProgressEvent;

implementation

{$R *.dfm}

procedure TfrmFx.DoProgress(Progress: Integer; Text: string; var Cancel: Boolean);
var p, w: Integer;
begin
  w:= pnlCtrls.Width - 16;
  if Progress = 0 then begin
    shpPrg.Width:= 1;
    shpPrg.Visible:= True;
  end
  else if Progress = 99 then shpPrg.Width:= w
  else if Progress < 100 then shpPrg.Width:= MulDiv(Progress, w, 100)
  else shpPrg.Visible:= False;
  shpPrg.Repaint;
end;

procedure fxSNN(Src: TIEBitmap; Radius: Integer);
var
  i, j, x, y, x1, y1, x2, y2, h, w, sumR, sumG, sumB, count, u, v, Row: Integer;
  d1, d2: Single;
  SL: array of PRGBArray;
  c: TRGB24;
  lr, lg, lb, r1, r2, g1, g2, b1, b2: Byte;
  DRow: PRGBArray;
  rmean, r, g, b: LongInt;
  bmp: TIEBitmap;

  P, Prg: Integer;
  DoPrg, Bool: Boolean;
begin
  DoPrg:= Assigned(fxProgress);
  Prg:= 0;
  if DoPrg then fxProgress(Prg, '', Bool);

  bmp:= TIEBitmap.Create;
  bmp.Assign(Src);

  h:= Src.Height - 1;
  w:= Src.Width - 1;

  DRow:= Src.ScanLine[0];
  Row:= Integer(Src.ScanLine[1]) - Integer(DRow);
  SetLength(SL, Bmp.Height);
  SL[0]:= Bmp.Scanline[0];
  i:= Integer(Bmp.Scanline[1]) - Integer(SL[0]);
  for j:= 1 to Bmp.Height - 1 do SL[j] := PRGBArray(Integer(SL[j - 1]) + i);

  count:= (Radius + 1 + Radius) * (Radius + 1 + Radius);
  for y:= 0 to h do begin
    for x:= 0 to w do begin
      sumR:= 0; sumG:= 0; sumB:= 0;
      c:= SL[y][x]; lr:= c.R; lg:= c.G; lb:= c.B;
      for v:= -Radius to Radius do begin
        for u:= -Radius to Radius do begin
          x1:= x + u; y1:= y + v;
          if x1 < 0 then x1:= 0 else if x1 > w then x1:= w;
          if y1 < 0 then y1:= 0 else if y1 > h then y1:= h;

          x2:= x - u; y2:= y - v;
          if x2 < 0 then x2:= 0 else if x2 > w then x2:= w;
          if y2 < 0 then y2:= 0 else if y2 > h then y2:= h;

          c:= SL[y1][x1]; r1:= c.R; g1:= c.G; b1:= c.B;
          c:= SL[y2][x2]; r2:= c.R; g2:= c.G; b2:= c.B;

          // Color distance formula is from:
          // http://www.compuphase.com/cmetric.htm
          rmean:= (lr + r1) shr 1;
          r:= lr - r1;
          g:= lg - g1;
          b:= lb - b1;
          d1:= Sqrt((((512+rmean)*r*r) shr 8) + 4*g*g + (((767-rmean)*b*b) shr 8));

          rmean:= (lr + r2) shr 1;
          r:= lr - r2;
          g:= lg - g2;
          b:= lb - b2;
          d2:= Sqrt((((512+rmean)*r*r) shr 8) + 4*g*g + (((767-rmean)*b*b) shr 8));

          if d1 < d2 then begin
            sumR:= sumR + r1;
            sumG:= sumG + g1;
            sumB:= sumB + b1;
          end else begin
            sumR:= sumR + r2;
            sumG:= sumG + g2;
            sumB:= sumB + b2;
          end;
        end;
      end;
      DRow[x].R:= Trunc(sumR / count);
      DRow[x].G:= Trunc(sumG / count);
      DRow[x].B:= Trunc(sumB / count);
    end;
    DRow:= PRGBArray(Integer(DRow) + Row);

    if DoPrg then begin
      p:= MulDiv(100, y, h);
      if p <> Prg then begin
        fxProgress(p, 'Median', Bool);
        Prg:= p;
      end;
    end;

  end;
  bmp.Free;
end;

procedure TfrmFx.DoFX;
begin
  fxProgress:= DoProgress;
  imgView.IEBitmap.Assign(TheImage);
  fxSNN(imgView.IEBitmap, tbRadius.Position);
  imgView.Update;
  fxProgress:= nil;
end;

procedure TfrmFx.tbRadiusChange(Sender: TObject);
begin
  lblRadius.Caption:= 'Radius(' + IntToStr(tbRadius.Position) + ' Pixels)';
end;

procedure TfrmFx.btnOpenClick(Sender: TObject);
begin
  if dlgOpen.Execute then begin
    imgView.IO.LoadFromFile(dlgOpen.Filename);
    TheImage.Assign(imgView.IEBitmap);
    TheImage.PixelFormat:= ie24RGB;
    imgView.IEBitmap.Assign(TheImage);
    ImgLoaded := True;
    imgView.Enabled := ImgLoaded;
  end;
end;

procedure TfrmFx.btnPreviewClick(Sender: TObject);
begin
  DoFX;
end;

procedure TfrmFx.btnSaveClick(Sender: TObject);
begin
  if dlgSave.Execute then imgView.IO.SaveToFile(dlgSave.FileName);
end;

procedure TfrmFx.FormCreate(Sender: TObject);
begin
  TheImage:= TIEBitmap.Create;
  imgView.IEBitmap.Width:= 1;
  imgView.IEBitmap.Height:= 1;
  imgView.IEBitmap.Canvas.Pixels[0, 0] := $00C4C4C4;
  imgView.MouseWheelParams.Action:= iemwNone;
end;

procedure TfrmFx.FormDestroy(Sender: TObject);
begin
  TheImage.Free;
end;

end.
